"""
Item Types API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> item_types_api = ItemTypesAPI(client)
    >>> item_types = item_types_api.get_item_types()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class ItemTypesAPI:
    client: JamaClient

    resource_path = "itemtypes"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_item_types(
        self,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        This method will return all item types of the across all projects of
        the Jama Connect instance.

        Args:
            allowed_results_per_page: Number of results per page

        Returns: An array of dictionary objects

        """
        return self.client.get_all(
            self.resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def get_item_type(
        self,
        item_type_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Gets item type information for a specific item type id.

        Args:
            item_type_id: The api id of the item type to fetch

        Returns: JSON object

        """
        resource_path = f"itemtypes/{item_type_id}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)



"""
Items API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> items_api = ItemsAPI(client)
    >>> items = items_api.get_items(project_id=1)
"""

import json
import logging
from typing import Optional

from py_jama_client.client import ClientResponse, JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException

py_jama_client_logger = logging.getLogger("py_jama_client")


class ItemsAPI:
    client: JamaClient

    resource_path = "items"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_items(
        self,
        project_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        This method will return all items in the specified project.
        Args:
            project_id: the project ID
            allowed_results_per_page: number of results per page

        Returns: a Json array of item objects

        """

        req_params = {"project": project_id}
        if params is None:
            params = req_params
        else:
            params.update(req_params)

        return self.client.get_all(
            self.resource_path,
            params=params,
            allowed_results_per_page=allowed_results_per_page,
        )

    def get_item(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        This method will return a singular item of a specified item id
        Args:
            item_id: the item id of the item to fetch

        Returns: a dictonary object representing the item

        """
        resource_path = f"{self.resource_path}/{item_id}"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def post_item(
        self,
        project_id: int,
        item_type_id: int,
        child_item_type_id: int,
        location: dict,
        fields: dict,
        global_id: int = None,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        This method will post a new item to Jama Connect.

        Args:
            project_id: (int) id of project
            item_type_id: (int) ID of an Item Type.
            child_item_type_id: (int) integer ID of an Item Type.
            location: (dict) containing key "parent"
            fields: (dict) dictionary item field data.
        Returns:
            newly created item

        "location": {
            "parent": {
            "item": 0,
                "project": 0
            }
        }
        """

        body = {
            "project": project_id,
            "itemType": item_type_id,
            "childItemType": child_item_type_id,
            "location": {"parent": location},
            "fields": fields,
        }
        resource_path = f"{self.resource_path}"

        # we setting a global ID?
        if global_id is not None:
            body["globalId"] = global_id
            params["setGlobalIdManually"] = True

        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def post_item_tag(
        self,
        item_id: int,
        tag_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ) -> int:
        """
        Add an existing tag to the item with the specified ID
        Args:
            item_id: The API ID of the item to add a tag.
            tag_id: The API ID of the tag to add to the item.

        Returns: 201 if successful

        """
        body = {"tag": tag_id}
        resource_path = f"{self.resource_path}/{item_id}/tags"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code

    def post_item_sync(
        self,
        source_item: int,
        pool_item: int,
        *args,
        params: Optional[dict],
        **kwargs,
    ):
        """
        add an item to an existing pool of global ids
        Args:
            source_item: integer API ID of the source item, this item will
                adopt the global id of the pool_item.
            pool_item: integer API ID of the item in the target global ID pool.

        Returns: the integer ID of the modified source item.
        """
        body = {"item": source_item}

        resource_path = f"{self.resource_path}/{pool_item}/synceditems"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def post_item_attachment(
        self,
        item_id: int,
        attachment_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ) -> int:
        """
        Add an existing attachment to the item with the specified ID
        :param item_id: this is the ID of the item
        :param attachment_id: The ID of the attachment
        :return: 201 if successful / the response status of the post operation
        """
        body = {"attachment": attachment_id}
        resource_path = f"items/{item_id}/attachments"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code

    def put_item(
        self,
        project_id: int,
        item_id: int,
        item_type_id: int,
        child_item_type_id: int,
        location: dict,
        fields: dict,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        This method will PUT a new item to Jama Connect.
        Args:
            project: integer representing the project to which this item is to
                be posted
            item_id: integer representing the item which is to be updated
            item_type_id: integer ID of an Item Type.
            child_item_type_id: integer ID of an Item Type.
            location: dictionary with a key of 'item' or 'project' and a value
                with the ID of the parent
            fields: dictionary item field data.
        Returns: integer ID of the successfully posted item or None if there
        was an error.
        """

        body = {
            "project": project_id,
            "itemType": item_type_id,
            "childItemType": child_item_type_id,
            "location": {"parent": location},
            "fields": fields,
        }
        resource_path = f"items/{item_id}"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        self.handle_response_status(response)
        return response.status_code

    def patch_item(
        self,
        item_id: int,
        patches: list[dict],
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ) -> int:
        """
        This method will patch an item.
        Args:
            item_id: the API ID of the item that is to be patched
            patches: An array of dicts, that represent patch operations each
                dict should have the following entries
             [
                {
                    "op": string,
                    "path": string,
                    "value": {}
                }
            ]

        Returns: The response status code

        """
        resource_path = f"items/{item_id}"
        headers = {"Content-Type": "application/json",
                   "Accept": "application/json"}

        try:
            response = self.client.patch(
                resource_path,
                params,
                data=json.dums(patches),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))

        JamaClient.handle_response_status(response)
        return response.status_code

    def delete_item(
        self,
        item_id: int,
        *args,
        **kwargs,
    ) -> int:
        """
        This method will delete an item in Jama Connect.

        Args:
            item_id: The jama connect API ID of the item to be deleted

        Returns: The success status code.
        """
        resource_path = f"items/{item_id}"
        try:
            response = self.client.delete(resource_path)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code

    def get_tagged_items(
        self,
        tag_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all items tagged with the specified ID

        Args:
            tag_id: The ID of the tag to fetch the results for.
            allowed_results_per_page: Number of results per page

        Returns:
            A List of items that match the tag.

        """
        resource_path = f"tags/{tag_id}/items"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def get_items_upstream_relationships(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Returns a list of all the upstream relationships for the item with the
        specified API ID.
        Args:
            item_id: the api id of the item
            allowed_results_per_page: number of results per page

        Returns: an array of dictionary objects that represent the upstream
        relationships for the item.

        """
        resource_path = "items/" + str(item_id) + "/upstreamrelationships"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def get_items_downstream_related(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Returns a list of all the downstream related items for the item with
        the specified API ID.

        Args:
            item_id: the api id of the item to fetch downstream items for
            allowed_results_per_page: number of results per page

        Returns: an array of dictionary objects that represent the downstream
        related items for the specified item.

        """
        resource_path = f"items/{item_id}/downstreamrelated"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def get_items_downstream_relationships(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Returns a list of all the downstream relationships for the item with
        the specified API ID.

        Args:
            item_id: the api id of the item

        Returns: an array of dictionary objects that represent the downstream
        relationships for the item.

        """
        resource_path = f"items/{item_id}/downstreamrelationships"
        return self.client.get_all(
            resource_path, allowed_results_per_page=allowed_results_per_page
        )

    def get_items_upstream_related(
        self, item_id: int, *args, params: Optional[dict] = None, **kwargs
    ):
        """
        Returns a list of all the upstream related items for the item with the
        specified API ID.

        Args:
            item_id: the api id of the item to fetch upstream items for

        Returns: an array of dictionary objects that represent the upstream
        related items for the specified item.

        """
        resource_path = f"items/{item_id}/upstreamrelated"
        return self.client.get_all(resource_path, params, **kwargs)

    def get_item_workflow_transitions(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all valid workflow transitions that can be made with the specified
        API ID

        Args:
            item_id: the api id of the item
            allowed_results_per_page: number of results per page

        Returns: an array of dictionary objects that represent the workflow
        transitions for the item.

        """
        resource_path = f"items/{item_id}/workflowtransitionoptions"
        return self.client.get_all(resource_path, params, **kwargs)

    def get_item_children(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        This method will return list of the child items of the item passed to
        the function.
        Args:
            item_id: (int) The id of the item for which children items should
            be fetched allowed_results_per_page: Number of results per page

        Returns: a List of Objects that represent the children of the item
        passed in.
        """
        resource_path = f"items/{item_id}/children"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def get_items_synceditems(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all synchronized items for the item with the specified ID

        Args:
            item_id: The API id of the item being
            allowed_results_per_page: Number of results per page

        Returns: A list of JSON Objects representing the items that are in the
        same synchronization group as the specified item.

        """
        resource_path = f"items/{item_id}/synceditems"
        return self.client.get_all(
            resource_path, allowed_results_per_page=allowed_results_per_page
        )

    def get_items_synceditems_status(
        self,
        item_id: int,
        synced_item_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the sync status for the synced item with the specified ID

        Args:
            item_id: The id of the item to compare against
            synced_item_id: the id of the item to check if it is in sync

        Returns: The response JSON from the API which contains a single field
            'inSync' with a boolean value.

        """
        resource_path = (
            f"items/{item_id}/synceditems/{synced_item_id}/syncstatus")
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_item_versions(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all versions for the item with the specified ID

        Args:
            item_id: the item id of the item to fetch

        Returns: JSON array with all versions for the item
        """
        resource_path = f"items/{item_id}/versions"
        return self.client.get_all(resource_path, params)

    def get_item_version(
        self,
        item_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the numbered version for the item with the specified ID

        Args:
            item_id: the item id of the item to fetch
            version_num: the version number for the item

        Returns: a dictionary object representing the numbered version
        """
        resource_path = f"items/{item_id}/versions/{version_num}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_versioned_item(
        self,
        item_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the snapshot of the item at the specified version

        Args:
            item_id: the item id of the item to fetch
            version_num: the version number for the item

        Returns: a dictionary object representing the versioned item
        """
        resource_path = f"items/{item_id}/versions/{version_num}/versioneditem"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_item_versions(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all versions for the item with the specified ID

        Args:
            item_id: the item id of the item to fetch
            allowed_results_per_page: number of results per page

        Returns: JSON array with all versions for the item
        """
        resource_path = f"items/{item_id}/versions"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page
        )

    def get_item_version(
        self,
        item_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the numbered version for the item with the specified ID

        Args:
            item_id: the item id of the item to fetch
            version_num: the version number for the item

        Returns: a dictionary object representing the numbered version
        """
        resource_path = f"items/{item_id}/versions/{version_num}"
        response = self.client.get(resource_path, params)
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_versioned_item(
        self,
        item_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the snapshot of the item at the specified version

        Args:
            item_id: the item id of the item to fetch
            version_num: the version number for the item

        Returns: a dictionary object representing the versioned item
        """
        resource_path = f"items/{item_id}/versions/{version_num}/versioneditem"
        response = self.client.get(resource_path, params)
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_item_lock(self, item_id: int, params: Optional[dict] = None):
        """
        Get the locked state, last locked date, and last locked by user for
        the item with the specified ID
        Args:
            item_id: The API ID of the item to get the lock info for.

        Returns:
            A JSON object with the lock information for the item with the
            specified ID.

        """
        resource_path = f"items/{item_id}/lock"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def put_item_lock(self, item_id: int, locked: bool) -> int:
        """
        Update the locked state of the item with the specified ID
        Args:
            item_id: the API id of the item to be updated
            locked: boolean lock state to apply to this item

        Returns:
            response status 200

        """
        body = {
            "locked": locked,
        }
        resource_path = f"items/{item_id}/lock"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path,
                data=json.dumps(body),
                headers=headers,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        return self.handle_response_status(response)

    def get_item_tags(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Return all tags for the item with the specified ID

        Args:
            item_id: the item id of the item to fetch
            allowed_results_per_page: number of results

        Returns: a dictionary object representing the item's tags

        """
        resource_path = f"items/{item_id}/tags"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
        )



__all__ = ["ClientResponse"]

from dataclasses import dataclass

from httpx import Response


@dataclass
class ClientResponse:
    meta: dict
    links: dict
    linked: dict
    data: list

    @classmethod
    def from_response(cls, response: Response):
        """
        Parse a response from the Jama API into a ClientResponse object.

        Args:
            response (httpx.Response): The response from the Jama API.

        Returns:
            ClientResponse: A ClientResponse object.
        """
        response_json: dict = response.json()

        return ClientResponse(
            meta=response_json.get("meta", {}),
            links=response_json.get("links", {}),
            linked=response_json.get("linked", {}),
            data=response_json.get("data", {}),
        )

    def to_dict(self):
        """
        Convert client response object to dictionary.
        """
        return {
            "meta": self.meta,
            "links": self.links,
            "linked": self.linked,
            "data": self.data,
        }

    def __add__(self, other):
        self.meta.update(other.meta)
        self.links.update(other.links)
        self.linked.update(other.linked)
        self.data += other.data
        return self



class CoreException(Exception):
    """This is the base class for all exceptions raised by the Core"""

    def __init__(self, message, status_code=None, reason=None):
        super(CoreException, self).__init__(message)
        self.status_code = status_code
        self.reason = reason


class UnauthorizedTokenException(CoreException):
    """This exception is thrown whenever fetching the oauth token returns a
    401 unauthorized response."""

    pass


class APIException(Exception):
    """This is the base class for all exceptions raised by the JamaClient"""

    def __init__(self, message, status_code=None, reason=None):
        super(APIException, self).__init__(message)
        self.status_code = status_code
        self.reason = reason


class UnauthorizedException(APIException):
    """This exception is thrown whenever the api returns a 401 unauthorized
    response."""

    pass


class TooManyRequestsException(APIException):
    """This exception is thrown whenever the api returns a 429 too many
    requests response."""

    pass


class ResourceNotFoundException(APIException):
    """This exception is raised whenever the api returns a 404 not found
    response."""

    pass


class AlreadyExistsException(APIException):
    """This exception is thrown when the API returns a 400 response with a
    message that the resource already exists."""

    pass


class APIClientException(APIException):
    """This exception is thrown whenever a unknown 400 error is encountered."""

    pass


class APIServerException(APIException):
    """This exception is thrown whenever an unknown 500 response is
    encountered."""

    pass



"""
Abstract Items API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> abstract_items_api = AbstractItemsAPI(client)
    >>> abstract_items = abstract_items_api.get_abstract_items()
"""

import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class AbstractItemsAPI:
    client: JamaClient

    resource_path = "abstractitems"

    def __init__(self, client: JamaClient) -> None:
        self.client = client

    def get_abstract_items(
        self,
        project: list[int] = None,
        item_type: list[int] = None,
        document_key: list[str] = None,
        release: list[int] = None,
        created_date: list[str] = None,
        modified_date: list[str] = None,
        last_activity_date: list[str] = None,
        contains: list[str] = None,
        sort_by: list[str] = None,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Search for items, test plans, test cycles, test runs, or attachments
        GET: /abstractitems/

        Args:
            project: list of project resource ids
            item_type: list of item type resource ids
            document_key: list of document keys
            release: list of release source ids
            created_date: Filter datetime fields after a single date or within a range of values.
                Provide one or two values in ISO8601 format (milliseconds or seconds) -
                "yyyy-MM-dd'T'HH:mm:ss.SSSZ" or "yyyy-MM-dd'T'HH:mm:ssZ"
            modified_date: Filter datetime fields after a single date or within a range of values.
                Provide one or two values in ISO8601 format (milliseconds or seconds) -
                "yyyy-MM-dd'T'HH:mm:ss.SSSZ" or "yyyy-MM-dd'T'HH:mm:ssZ"
            last_activity_date: Filter datetime fields after a single date or within a range of values.
                Provide one or two values in ISO8601 format (milliseconds or seconds) -
                "yyyy-MM-dd'T'HH:mm:ss.SSSZ" or "yyyy-MM-dd'T'HH:mm:ssZ"
            contains: Filter on the text contents of the item. Strings taken literally.
                Multiple 'contains' values will be bitwise ORed.
            sort_by: Sort orders can be added with the name of the field by which to sort, followed by .asc
                or .desc (e.g. 'name.asc' or 'modifiedDate.desc'). If not set, this defaults to sorting
                by sequence.asc and then documentKey.asc
        """

        # Add each parameter that is not null to the request.
        if params is None:
            params = {}

        if project is not None:
            params.update({"project": project})

        if item_type is not None:
            params.update({"itemType": item_type})

        if document_key is not None:
            params.update({"documentKey": document_key})

        if release is not None:
            params.update({"release": release})

        if created_date is not None:
            params.update({"createdDate": created_date})

        if modified_date is not None:
            params.update({"modifiedDate": modified_date})

        if last_activity_date is not None:
            params.update({"lastActivityDate": last_activity_date})

        if contains is not None:
            params.update({"contains": contains})

        if sort_by is not None:
            params.update({"sortBy": sort_by})

        return self.client.get_all(
            self.resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )

    def get_abstract_item(
        self,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get any item, test plan, test cycle, test run, or attachment with the specified ID
        GET: /abstractitems/{item_id}

        Args:
            item_id: the item id of the item to fetch
        """
        resource_path = f"abstractitems/{item_id}"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_abstract_versioned_relationships(
        self,
        item_id: int,
        timestamp: str,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all versioned relationships that were associated to the item at the specified time
        GET: /abstractitems/{item_id}/versionedrelationships

        Args:
            item_id: item resource id
            timestamp: Get relationships for the specified item at this date and time.
                Requires ISO8601 formatting (milliseconds or seconds) - "yyyy-MM-dd'T'HH:mm:ss.SSSZ"
                or "yyyy-MM-dd'T'HH:mm:ssZ"
        """
        resource_path = f"{self.resource_path}/{item_id}/versionedrelationships"
        req_params = {"timestamp": timestamp}
        if params is None:
            params = req_params
        else:
            params.update(req_params)
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )

    def get_abstract_item_versions(
        self,
        item_id: int,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all versions for the item with the specified ID
        GET: /abstractitems/{item_id}/versions

        Args:
            item_id: the item id of the item to fetch

        Returns: JSON array with all versions for the item
        """
        resource_path = f"{self.resource_path}/{item_id}/versions"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )

    def get_abstract_item_version(
        self,
        item_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the numbered version for the item with the specified ID
        GET: /abstractitems/{item_id}/versions/{version_num}/

        Args:
            item_id: the item id of the item to fetch
            version_num: the version number for the item
        """
        resource_path = f"{self.resource_path}/{item_id}/versions/{version_num}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_abstract_versioned_item(
        self,
        item_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the snapshot of the item at the specified version
        GET: /abstractitems/{item_id}/versions/{version_num}/versioneditem/

        Args:
            item_id: the item id of the item to fetch
            version_num: the version number for the item
        """
        resource_path = f"{self.resource_path}/{item_id}/versions/{version_num}/versioneditem"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)



"""
Attachments API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> attachments_api = AttachmentsAPI(client)
    >>> attachments = attachments_api.get_attachment(attachment_id=10)
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class AttachmentsAPI:
    client: JamaClient

    resource_path = "attachments"

    def __init__(self, client: JamaClient) -> None:
        self.client = client

    def get_attachment(
        self,
        attachment_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the attachment with the specified ID
        GET: /attachments/{attachmentId}/

        Args:
            attachment_id: the attachment id of the attachment to fetch
        """
        resource_path = f"{self.resource_path}/{attachment_id}"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_attachment_file(
        self,
        attachment_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ) -> bytes:
        """
        Download attachment file from the attachment with the specified ID
        GET: /attachments/{attachmentId}/file/

        Args:
            attachment_id: attachment resource id
        """
        resource_path = "files"
        req_params = {"url": attachment_id}

        if params is None:
            params = req_params
        else:
            params.update(req_params)

        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.content

    def put_attachments_file(
        self,
        attachment_id: int,
        file_path: str,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ) -> int:
        """
        Upload attachment file to the attachment with the specified ID
        PUT: /attachments/{attachmentId}/file

        Args:
            attachment_id: the integer ID of the attachment item to which we
                are uploading the file
            file_path: the file path of the file to be uploaded
        """
        resource_path = f"attachments/{attachment_id}/file"
        with open(file_path, "rb") as f:
            files = {"file": f}
            try:
                response = self.client.put(
                    resource_path,
                    params,
                    files=files,
                    **kwargs,
                )
            except CoreException as err:
                py_jama_client_logger.error(err)
                raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code

    def get_attachment_lock(
        self,
        attachment_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the locked state, last locked date, and last locked by user for
        the item with the specified ID
        GET: /attachments/{attachmentId}/lock

        Args:
            attachment_id: attachment resource id
        """
        resource_path = f"{self.resource_path}/{attachment_id}/lock"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def put_attachment_lock(
        self,
        attachment_id: int,
        locked: bool,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Update the locked state of the item with the specified ID
        PUT: /attachments/{attachmentId}/lock

        Args:
            attachment_id: attachment resource id
            locked: (bool) locked state
        """
        resource_path = f"{self.resource_path}/{attachment_id}/lock"
        body = {"locked": locked}
        try:
            response = self.client.put(
                resource_path, params, data=json.dumps(body), **kwargs
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_attachment_versions(
        self,
        attachment_id: int,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all versions for the item with the specified ID
        GET: /attachments/{attachmentId}/versions/

        Args:
            attachment_id: attachment resource id
        """
        resource_path = f"{self.resource_path}/{attachment_id}/versions/"
        return self.client.get_all(
            resource_path, params, allowed_results_per_page, **kwargs
        )

    def get_attachment_version(
        self,
        attachment_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the numbered version for the item with the specified ID
        GET: /attachments/{attachmentId}/versions/{versionNum}
        """
        resource_path = (
            f"{self.resource_path}/{attachment_id}/versions/{version_num}")
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_attachment_version_item(
        self,
        attachment_id: int,
        version_num: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the snapshot of the attachment at the specified version
        GET: /attachments/{attachmentId}/versions/{versionNum}/versionedItem
        """
        resource_path = (
            f"{self.resource_path}/{attachment_id}/versions/"
            f"{version_num}/versionedItem"
        )
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)



"""
Core Jama Connect Client API class

This module contains core classes for interacting with the Jama Connect API
"""

__all__ = ["JamaClient"]

import json
import logging
import math
import ssl
import time
import typing
from typing import Optional, Tuple

import httpx
import urllib3
from httpx import Response

from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import (AlreadyExistsException,
                                       APIClientException, APIException,
                                       APIServerException, CoreException,
                                       ResourceNotFoundException,
                                       TooManyRequestsException,
                                       UnauthorizedException,
                                       UnauthorizedTokenException)
from py_jama_client.response import ClientResponse

__DEBUG__ = False

# disable warnings for ssl verification
urllib3.disable_warnings()

py_jama_client_logger = logging.getLogger("py_jama_client")


class JamaClient:
    """
    Base client class
    """

    session_class = httpx.Client

    def __init__(
        self,
        host: str,
        credentials: Tuple[str, str] = ("username|client_id",
                                        "password|client_secret"),
        api_version: str = "/rest/v1/",
        oauth: bool = False,
        verify: typing.Union[bool, str, ssl.SSLContext] = True,
        timeout: int = 30,
    ):
        """
        Args:
            host: The host name of the Jama Connect server
            credentials: A tuple of username and password or client_id and
                client_secret
            api_version: The version of the Jama Connect API to use
            oauth: A boolean to indicate if OAuth is being used. Must be set
                to true if credentials are client id/secret pair.
            verify: SSL certificates (a.k.a CA bundle) used to verify the
                identity of requested hosts. Either True (default CA bundle),
                a path to an SSL certificate file, an ssl.SSLContext, or False.
            timeout: The timeout for the session
        """
        # Instance variables
        self.__api_version = api_version
        self.__host_name = host + self.__api_version
        self.__credentials = credentials
        self.__oauth = oauth
        self.__verify = verify
        self.__session = self.session_class(verify=verify, timeout=timeout)

        # Setup OAuth if needed.
        if self.__oauth:
            self.__token_host = host + "/rest/oauth/token"
            self.__token = None
            self.__get_fresh_token()

    def get_available_endpoints(self):
        try:
            response = self.__session.get("")
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def close(self) -> None:
        """Method to close underlying session"""
        self.__session.close()

    async def close(self) -> None:
        raise RuntimeError("await syntax not supported on sync core client"
                           "class")

    def delete(self, resource: str, **kwargs):
        """This method will perform a delete operation on the specified
        resource"""
        url = self.__host_name + resource

        if self.__oauth:
            self.__check_oauth_token()
            kwargs["headers"] = self.__add_auth_header(**kwargs)
            return self.__session.delete(url, **kwargs)

        return self.__session.delete(url, auth=self.__credentials, **kwargs)

    def get(self, resource: str, params: dict = None, **kwargs):
        """This method will perform a get operation on the specified
        resource"""
        url = self.__host_name + resource

        if self.__oauth:
            self.__check_oauth_token()
            kwargs["headers"] = self.__add_auth_header(**kwargs)
            return self.__session.get(url, params=params, **kwargs)

        return self.__session.get(url,
                                  auth=self.__credentials,
                                  params=params,
                                  **kwargs)

    def patch(self,
              resource: str,
              params: dict = None,
              data=None,
              json=None,
              **kwargs):
        """This method will perform a patch operation to the specified
        resource"""
        url = self.__host_name + resource

        if self.__oauth:
            self.__check_oauth_token()
            kwargs["headers"] = self.__add_auth_header(**kwargs)
            return self.__session.patch(
                url, params=params, data=data, json=json, **kwargs
            )

        return self.__session.patch(
            url,
            auth=self.__credentials,
            params=params,
            data=data,
            json=json,
            **kwargs
        )

    def post(self,
             resource: str,
             params: dict = None,
             data=None,
             json=None,
             **kwargs):
        """This method will perform a post operation to the specified
        resource."""
        url = self.__host_name + resource

        if self.__oauth:
            self.__check_oauth_token()
            kwargs["headers"] = self.__add_auth_header(**kwargs)
            return self.__session.post(
                url, params=params, data=data, json=json, **kwargs
            )

        return self.__session.post(
            url,
            auth=self.__credentials,
            params=params,
            data=data,
            json=json,
            **kwargs
        )

    def put(self,
            resource: str,
            params: dict = None,
            data=None,
            json=None,
            **kwargs):
        """This method will perform a put operation to the specified
        resource"""
        url = self.__host_name + resource

        if self.__oauth:
            self.__check_oauth_token()
            kwargs["headers"] = self.__add_auth_header(**kwargs)
            return self.__session.put(
                url, data=data, params=params, json=json, **kwargs
            )

        return self.__session.put(
            url,
            auth=self.__credentials,
            data=data,
            params=params,
            json=json,
            **kwargs
        )

    def __check_oauth_token(self):
        if self.__token is None:
            self._get_fresh_token()

        else:
            time_elapsed = time.time() - self.__token_acquired_at
            time_remaining = self.__token_expires_in - time_elapsed
            if time_remaining < 60:
                # if less than a minute remains, just get another token.
                self.__get_fresh_token()

    def __get_fresh_token(self):
        """This method will fetch a new oauth bearer token from the oauth
        token server."""
        data = {"grant_type": "client_credentials"}

        # By getting the system time before we get the token we avoid a
        # potential bug where the token may be expired.
        time_before_request = time.time()

        # Post to the token server, check if authorized
        try:
            response = httpx.post(
                self.__token_host,
                auth=self.__credentials,
                data=data,
                verify=self.__verify,
            )
            response.raise_for_status()
        except httpx.HTTPError as err:
            raise UnauthorizedTokenException(f"Unable to fetch token: {err}")

        # If success get relevant data
        if response.status_code in [200, 201]:
            response_json = response.json()
            self.__token = response_json["access_token"]
            self.__token_expires_in = response_json["expires_in"]
            self.__token_acquired_at = math.floor(time_before_request)

        else:
            py_jama_client_logger.error("Failed to retrieve OAuth Token")

    def __add_auth_header(self, **kwargs):
        headers = kwargs.get("headers")
        if headers is None:
            headers = {}
        headers["Authorization"] = "Bearer " + self.__token
        return headers

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.__session.close()

    def get_all(
        self,
        resource,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        This method will get all of the resources specified by the resource
        parameter, if an id or some other parameter is required for the
        resource, include it in the params parameter.
        Returns a single JSON array with all of the retrieved items.
        """

        if allowed_results_per_page < 1 or allowed_results_per_page > 50:
            raise ValueError("Allowed results per page must be between 1 "
                             "and 50")

        start_index = 0
        allowed_results_per_page = 20
        total_results = float("inf")

        data, meta, links, linked = [], {}, {}, {}
        while len(data) < total_results:
            page = self.get_page(resource,
                                 start_index,
                                 params=params,
                                 **kwargs)

            meta.update(page.meta)
            links.update(page.links)

            for item_type_key in page.linked:
                if item_type_key not in linked:
                    linked[item_type_key] = {}
                linked[item_type_key] = {
                    **linked[item_type_key],
                    **page.linked[item_type_key],
                }

            page_info = page.meta.get("pageInfo")
            start_index = (page_info.get("startIndex") +
                           allowed_results_per_page)
            total_results = page_info.get("totalResults")
            data.extend(page.data)

        return ClientResponse(meta, links, linked, data)

    def get_page(
        self,
        resource,
        start_at,
        params: dict = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        This method will return one page of results from the specified
        resource type.
        Pass any needed parameters along
        The response object will be returned
        """
        pagination = {"startAt": start_at,
                      "maxResults": allowed_results_per_page}

        if params is None:
            params = pagination
        else:
            params.update(pagination)

        try:
            response = self.get(resource, params=params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    @staticmethod
    def handle_response_status(response: Response):
        """
        Utility method for checking http status codes.
        If the response code is not in the 200 range, An exception will be
        thrown.
        """

        status = response.status_code

        if status in range(200, 300):
            return status

        if status in range(400, 500):
            """These are client errors. It is likely that something is wrong
            with the request."""

            response_message = "No Response"

            try:
                response_json = json.loads(response.text)
                response_message = response_json.get("meta").get("message")

            except json.JSONDecodeError:
                pass

            # Log the error
            py_jama_client_logger.error(
                "API Client Error. Status: {} Message: {}".format(
                    status, response_message
                )
            )

            if (response_message is not None and
               "already exists" in response_message):
                raise AlreadyExistsException(
                    "Entity already exists.",
                    status_code=status,
                    reason=response_message,
                )

            if status == 401:
                raise UnauthorizedException(
                    "Unauthorized: check credentials and permissions.  "
                    "API response message {}".format(response_message),
                    status_code=status,
                    reason=response_message,
                )

            if status == 404:
                raise ResourceNotFoundException(
                    "Resource not found. check host url.",
                    status_code=status,
                    reason=response_message,
                )

            if status == 429:
                raise TooManyRequestsException(
                    "Too many requests.  API throttling limit reached, or "
                    "system under maintenance.",
                    status_code=status,
                    reason=response_message,
                )

            raise APIClientException(
                "{} {} Client Error.  Bad Request.  "
                "API response message: {}".format(
                    status, response.reason, response_message
                ),
                status_code=status,
                reason=response_message,
            )

        if status in range(500, 600):
            """These are server errors and network errors."""

            # Log The Error
            py_jama_client_logger.error(
                "{} Server error. {}".format(status, response.reason)
            )
            raise APIServerException(
                "{} Server Error.".format(status),
                status_code=status,
                reason=response.reason,
            )

        # Catch anything unexpected
        py_jama_client_logger.error("{} error. {}".format(status,
                                                          response.reason))
        raise APIException(
            "{} error".format(status),
            status_code=status,
            reason=response.reason
        )



"""
Filters API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> filters_api = FiltersAPI(client)
    >>> filters = filters_api.get_filter_results()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class FiltersAPI:
    client: JamaClient

    resource_path = "testruns"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_filter_results(
        self,
        filter_id: int,
        project_id: int = None,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all results items for the filter with the specified ID

        Args:
            filter_id: The ID of the filter to fetch the results for.
            project_id: Use this only for filters that run on any project,
                where projectScope is CURRENT
            allowed_results_per_page: Number of results per page

        Returns:
            A List of items that match the filter.

        """
        resource_path = f"filters/{filter_id}/results"

        req_params = {"project": project_id}

        if params is None:
            params = req_params
        else:
            params.update(req_params)

        return self.client.get_all(
            resource_path,
            params=params,
            allowed_results_per_page=allowed_results_per_page,
        )



"""
Test Plans API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> test_plans_api = TestPlansAPI(client)
    >>> test_plans = test_plans_api.get_test_plans()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class TestPlansAPI:
    client: JamaClient

    resource_path = "testplans"

    def __init__(self, client: JamaClient):
        self.client = client

    def post_testplans_testcycles(
        self,
        testplan_id: int,
        testcycle_name: str,
        start_date: str,
        end_date: str,
        testgroups_to_include: list[int] = None,
        testrun_status_to_include: list[str] = None,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        This method will create a new Test Cycle.

        Args:
            testplan_id (int): The API_ID of the testplan to create the test
                cycle from.
            testcycle_name (str): The name you would like to set for the new
                Test Cycle
            start_date (str): Start date in 'yyyy-mm-dd' Format
            end_date (str): End date in 'yyyy-mm-dd' Format
            testgroups_to_include (int[]):  This array of integers specify the
                test groups to be included.
            testrun_status_to_include (str[]): Only valid after generating the
                first Test Cycle, you may choose to only generate Test Runs
                that were a specified status in the previous cycle. Do not
                specify anything to include all statuses

        Returns:
            (int): Returns the the newly created testcycle
        """
        resource_path = f"testplans/{testplan_id}/testcycles"
        headers = {"content-type": "application/json"}
        fields = {"name": testcycle_name,
                  "startDate": start_date,
                  "endDate": end_date}
        test_run_gen_config = {}
        if testgroups_to_include is not None:
            test_run_gen_config["testGroupsToInclude"] = testgroups_to_include
        if testrun_status_to_include is not None:
            test_run_gen_config["testRunStatusesToInclude"] = (
                testrun_status_to_include)
        body = {"fields": fields,
                "testRunGenerationConfig": test_run_gen_config}

        # Make the API Call
        try:
            response = self.client.post(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))

        # Validate response
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)



"""
Test Runs API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> test_runs_api = TestRunsAPI(client)
    >>> test_runs = test_runs_api.get_test_runs()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class TestRunsAPI:
    client: JamaClient

    resource_path = "testruns"

    def __init__(self, client: JamaClient):
        self.client = client

    def put_test_run(
        self,
        test_run_id: int,
        data: dict = None,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """This method will post a test run to Jama through the API"""
        resource_path = f"testruns/{test_run_id}"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path,
                params,
                data=data,
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code



"""
Pick Lists API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> pick_lists_api = PickListsAPI(client)
    >>> pick_lists = pick_lists_api.get_pick_lists()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class PickListsAPI:
    client: JamaClient

    resource_path = "picklists"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_pick_lists(
        self,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Returns a list of all the pick lists

        Args:
            allowed_results_per_page: number of results per page

        Returns: an array of dictionary objects
        """
        return self.client.get_all(
            self.resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
        )

    def get_pick_list(
        self,
        pick_list_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Gets all a singular picklist

        Args:
            pick_list_id: The API id of the pick list to fetch

        Returns: a dictionary object representing the picklist.
        """
        resource_path = f"picklists/{pick_list_id}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_pick_list_options(
        self,
        pick_list_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Gets all all the picklist options for a single picklist
        Args:
            pick_list_id: the api id of the picklist to fetch options for.
            allowed_results_per_page: number of results per page

        Returns: an array of dictionary objects that represent the
        picklist options.
        """
        resource_path = f"picklists/{pick_list_id}/options"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page
        )



"""
Projects API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> projects_api = ProjectsAPI(client)
    >>> projects = projects_api.get_projects()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import ClientResponse, JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import (APIException, CoreException,
                                       ResourceNotFoundException)

py_jama_client_logger = logging.getLogger("py_jama_client")


class ProjectsAPI:
    client: JamaClient

    resource_path = "projects/"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_projects(
        self,
        params: Optional[dict] = None,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
    ):
        """
        This method will return all projects as JSON object
        optional: if project_id is specified, it will return a single project
        Args:
            allowed_results_per_page: number of results per page
        Returns:
            JSON Array of Item Objects.
        """
        resource_path = "projects"

        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page
        )

    def get_project_by_id(self,
                          project_id: int,
                          params: Optional[dict] = None):
        """
        This method will return a single project as JSON object
        Args:
            project_id: the id of the project to fetch

        Returns: a dictionary object representing the project

        """
        resource_path = f"projects/{project_id}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise ResourceNotFoundException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_relationship_rule_set_projects(self, id: int):
        """
        This method will return the projects that have a given relationship
        rule set defined.

        Returns: An array of the dictionary objects representing the projects
        with a given rule set assigned

        """
        resource_path = f"relationshiprulesets/{id}/projects"
        return self.client.get_all(resource_path)

    def post_project_attachment(
        self,
        project_id: int,
        name: str,
        description: str,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        This Method will make a new attachment object in the specified project
        :param project_id: The integer project ID to create the attachment in.
        :param name:  The name of the attachment
        :param description: The description of the attachment
        :return: Returns the ID of the newly created attachment object.
        """
        body = {"fields": {"name": name, "description": description}}

        resource_path = f"projects/{project_id}/attachments"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def put_project_item_type(
        self,
        project_id: int,
        item_type_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Add item type to project
        Args:
            :param project_id: The integer project ID to add item type to.
            :param item_type_id integer ID of an Item Type.

        Returns:
            response status 200

        """
        resource_path = f"projects/{project_id}/itemtypes/{item_type_id}"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path, params, headers=headers, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code



"""
Pick List Options API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> pick_list_options_api = PickListOptionsAPI(client)
    >>> pick_list_options = pick_list_options_api.get_pick_list_options(
            pick_list_id=10)
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class PickListOptionsAPI:
    client: JamaClient

    resource_path = "picklistoptions"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_pick_list_option(
        self,
        pick_list_option_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Fetches a single picklist option from the API
        Args:
            pick_list_option_id: The API ID of the picklist option to fetch

        Returns: A dictonary object representing the picklist option.

        """
        resource_path = f"picklistoptions/{pick_list_option_id}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)



"""
Test Cycles API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> test_cycles_api = TestCyclesAPI(client)
    >>> test_cycles = test_cycles_api.get_test_cycles()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class TestCyclesAPI:
    client: JamaClient

    resource_path = "testcycles"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_test_cycle(
        self,
        test_cycle_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        This method will return JSON data about the test cycle specified by
        the test cycle id.

        Args:
            test_cycle_id: the api id of the test cycle to fetch

        Returns: a dictionary object that represents the test cycle

        """
        resource_path = f"testcycles/{test_cycle_id}"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_test_cycle_runs(
        self,
        test_cycle_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        This method will return all test runs associated with the specified
        test cycle.  Test runs will be returned as a list of json objects.
        Args:
            test_cycle_id: (int) The id of the test cycle
        """
        resource_path = f"testcycles/{test_cycle_id}/testruns"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )



"""
Baselines API module

Example usage:
    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> baselines_api = BaselinesAPI(client)
    >>> baselines = baselines_api.get_baselines(project_id=82)
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class BaselinesAPI:
    client: JamaClient

    resource_path = "baselines"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_baselines(
        self,
        project_id: int,
        *args,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        params: Optional[dict] = None,
        **kwargs,
    ) -> ClientResponse:
        """
        Get all baselines in the project with the specified ID
        GET: /baselines/
        Args:
            project_id:  project resource id
        """
        req_params = {"project": project_id}
        if params is None:
            params = req_params
        else:
            params.update(req_params)

        return self.client.get_all(
            self.resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )

    def get_baseline(
        self,
        baseline_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the baseline with the specified ID
        GET: /baselines/{baselineId}/

        Args:
            baseline_id: the id of the baseline to fetch
        """
        resource_path = f"{self.resource_path}/{baseline_id}"

        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def put_baseline(
        self,
        baseline_id: int,
        source: int,
        baseline_origin_type: str,
        baseline_origin_id: str,
        name: str,
        description: str,
        baseline_status_pick_list_option: int,
        *args,
        **kwargs,
    ):
        """
        Update the baseline with the specified ID
        PUT: /baselines/{baselineId}/

        Args:
            baseline_id: baseline resource id
        """
        resource_path = f"{self.resource_path}/{baseline_id}"
        body = {
            "source": source,
            "baselineOriginType": baseline_origin_type,
            "baselineOriginId": baseline_origin_id,
            "name": name,
            "description": description,
            "baselineStatusPickListOption": baseline_status_pick_list_option,
        }
        try:
            response = self.client.put(resource_path,
                                       data=json.dumps(body) ** kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def delete_baseline(
        self,
        baseline_id: int,
        *args,
        **kwargs,
    ):
        """
        Delete the baseline with the specified ID
        DELETE: /baselines/{baselineId}

        Args:
            baseline_id: baseline resource id
        """
        resource_path = f"{self.resource_path}/{baseline_id}"
        try:
            response = self.client.delete(resource_path, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
            return JamaClient.handle_response_status(response)

    def get_baseline_review_link(
        self,
        baseline_id: int,
        *args,
        **kwargs,
    ):
        """
        Get related review link
        GET: /baselines/{baselineId}/reviewlink

        Args:
            baseline_id: baseline resource id
        """
        resource_path = f"{self.resource_path}/{baseline_id}/reviewlink"
        try:
            response = self.client.get(resource_path, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_baseline_versioned_items(
        self,
        baseline_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all baseline items in a baseline with the specified ID
        Args:
            baseline_id:  baseline resource id
        """
        resource_path = f"baselines/{baseline_id}/versioneditems"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )

    def get_baseline_versioned_item(
        self,
        baseline_id: int,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the baseline item with the specified ID in a baseline with the
        specified ID
        GET: /baselines/{baselineId}/versioneditems/{itemId}

        Args:
            baseline_id: baseline resource id
            item_id: baseline item resource id
        """
        resource_path = (
            f"{self.resource_path}/{baseline_id}/versioneditems/{item_id}")
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_baseline_versioned_item_relationships(
        self,
        baseline_id: int,
        item_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all versioned relationships for the item in the baseline
        GET:
         /baselines/{baselineId}/versioneditems/{itemId}/versionedrelationships

        Args:
            baseline_id: baseline resource id
            item_id: baseline item resource id
        """
        resource_path = (
            f"{self.resource_path}/{baseline_id}/versioneditems/"
            f"{item_id}/versionedrelationships")
        return self.client.get_all(
            resource_path, params, allowed_results_per_page, **kwargs
        )



"""
Relationships API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> relationships_api = RelationshipsAPI(client)
    >>> relationships = relationships_api.get_relationships(project_id=82)
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class RelationshipsAPI:
    client: JamaClient

    resource_path = "relationships"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_relationships(
        self,
        project_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Returns a list of all relationships of a specified project

        Args:
            project_id: the api project id of a project
            allowed_results_per_page: number of results per page

        Returns: a list of dictionary objects that represents a relationships

        """
        resource_path = "relationships"

        req_params = {"project": project_id}
        if params is None:
            params = req_params
        else:
            params.update(req_params)

        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
        )

    def get_relationship(
        self,
        relationship_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Returns a specific relationship object of a specified relationship ID

        Args:
            relationship_id: the api project id of a relationship

        Returns: a dictionary object that represents a relationship

        """
        resource_path = f"relationships/{relationship_id}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def post_relationship(
        self,
        from_item: int,
        to_item: int,
        relationship_type: int = None,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Args:
            from_item: integer API id of the source item
            to_item: integer API id of the target item
            relationship_type: Optional integer API id of the relationship
                type to create

        Returns: The integer ID of the newly created relationship.

        """
        body = {
            "fromItem": from_item,
            "toItem": to_item,
        }
        if relationship_type is not None:
            body["relationshipType"] = relationship_type
        resource_path = "relationships/"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def put_relationship(
        self,
        relationship_id: int,
        from_item: int,
        to_item: int,
        relationship_type: int = None,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Args:
            relationship_id: integer API id of the relationship
            from_item: integer API id of the source item
            to_item: integer API id of the target item
            relationship_type: Optional integer API id of the relationship
                type to create
        """
        body = {"fromItem": from_item, "toItem": to_item}
        if relationship_type is not None:
            body["relationshipType"] = relationship_type
        resource_path = "relationships/{}".format(relationship_id)
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def delete_relationship(self, relationship_id: int) -> int:
        """
        Deletes a relationship with the specified relationship ID

        Args:
            relationship_id: the api project id of a relationship

        Returns: The success status code.

        """
        resource_path = f"relationships/{relationship_id}"
        try:
            response = self.client.delete(resource_path)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code

    def delete_relationship_suspect(self, relationship_id: int) -> int:
        """
        Removes suspect status of a relationship with the
        specified relationship ID

        Args:
            relationship_id: the api project id of a relationship

        Returns: The success status code.

        """
        resource_path = f"relationships/{relationship_id}/suspect"
        try:
            response = self.client.delete(resource_path)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code

    def get_relationship_rule_sets(self):
        """
        This method will return all relationship rule sets across all projects
        of the Jama Connect instance.

        Returns: An array of dictionary objects representing a rule set and
        its associated rules

        """
        resource_path = "relationshiprulesets/"
        return self.client.get_all(resource_path)

    def get_relationship_rule_set(self, relationship_id: int):
        """
        This method will return the relationship rule sets by id.

        Returns: A dictionary object representing a rule set and its
                 associated rules

        """
        resource_path = f"relationshiprulesets/{relationship_id}"
        response = self.client.get(resource_path)
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_relationship_types(
        self,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        This method will return all relationship types of the across all
        projects of the Jama Connect instance.

        Args:
            allowed_results_per_page: Number of results per page

        Returns: An array of dictionary objects

        """
        resource_path = "relationshiptypes/"
        return self.client.get_all(
            resource_path, allowed_results_per_page=allowed_results_per_page
        )

    def get_relationship_type(
        self,
        relationship_type_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs
    ):
        """
        Gets relationship type information for a specific relationship type id.

        Args:
            relationship_type_id: The api id of the item type to fetch

        Returns: JSON object

        """
        resource_path = f"relationshiptypes/{relationship_type_id}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)



"""
User API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> users_api = UsersAPI(client)
    >>> users = users_api.get_users()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class UsersAPI:
    client: JamaClient

    resource_path = "users"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_users(
        self,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Gets a list of all active users visible to the current user

        Args:
            allowed_results_per_page: Number of results per page

        Returns: JSON array

        """
        return self.client.get_all(
            self.resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def get_user(
        self,
        user_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Gets a single speificed user

        Args:
            user_id: user api ID

        Returns: JSON obect

        """
        resource_path = f"{self.resource_path}/{user_id}"
        try:
            response = self.client.get(
                resource_path,
                params,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        return ClientResponse.from_response(response)

    def get_current_user(
        self,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Gets a current user

        Returns: JSON obect

        """
        resource_path = f"{self.resource_path}/current"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        return ClientResponse.from_response(response)

    def get_current_user_favorite_filters(
        self,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Gets a list of favorite filters for the current user

        Args:
            allowed_results_per_page: Number of results per page

        Returns: JSON array

        """
        resource_path = f"{self.resource_path}/current/favoritefilters"

        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def post_user(
        self,
        username: str,
        password: str,
        first_name: str,
        last_name: str,
        email: str,
        license_type: str,
        phone: str = None,
        title: str = None,
        location: str = None,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Creates a new user

        Args:
            username: str
            password: str
            first_name: str
            last_name: str
            email: str
            phone: str - optional
            title: str - optional
            location: str - optional
            licenseType: enum [ NAMED, FLOATING, STAKEHOLDER,
                                FLOATING_COLLABORATOR,
                                RESERVED_COLLABORATOR, FLOATING_REVIEWER,
                                RESERVED_REVIEWER, NAMED_REVIEWER,
                                TEST_RUNNER, EXPIRING_TRIAL, INACTIVE ]

        Returns: newly created user

        """

        body = {
            "username": username,
            "password": password,
            "firstName": first_name,
            "lastName": last_name,
            "email": email,
            "phone": phone,
            "title": title,
            "location": location,
            "licenseType": license_type,
        }
        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                self.resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def put_user(
        self,
        user_id: int,
        username: str,
        password: str,
        first_name: str,
        last_name: str,
        email: str,
        phone: str = None,
        title: str = None,
        location: str = None,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        updates an existing user

        Args:
            username: str
            password: str
            first_name: str
            last_name: str
            email: str
            phone: str - optional
            title: str - optional
            location: str - optional

        Returns: api status code

        """

        body = {
            "username": username,
            "password": password,
            "firstName": first_name,
            "lastName": last_name,
            "email": email,
            "phone": phone,
            "title": title,
            "location": location,
        }
        resource_path = f"{self.resource_path}/{user_id}"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        self.handle_response_status(response)
        return response.status_code

    def put_user_active(
        self,
        user_id: int,
        is_active: bool,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        updates an existing users active status

        Args:
            is_active: boolean

        Returns: api status code

        """
        body = {"active": is_active}
        resource_path = f"{self.resource_path}/{user_id}/active"
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        self.handle_response_status(response)
        return response.status_code



from .abstract_items_api import AbstractItemsAPI
from .activities_api import ActivitiesAPI
from .attachments_api import AttachmentsAPI
from .baselines_api import BaselinesAPI
from .filters_api import FiltersAPI
from .item_types_api import ItemTypesAPI
from .items_api import ItemsAPI
from .pick_list_options_api import PickListOptionsAPI
from .pick_lists_api import PickListsAPI
from .projects_api import ProjectsAPI
from .relationships_api import RelationshipsAPI
from .tags_api import TagsAPI
from .test_cycles_api import TestCyclesAPI
from .test_plans_api import TestPlansAPI
from .test_runs_api import TestRunsAPI
from .users_api import UsersAPI

__all__ = [
    "AbstractItemsAPI",
    "ActivitiesAPI",
    "AttachmentsAPI",
    "BaselinesAPI",
    "FiltersAPI",
    "ItemTypesAPI",
    "ItemsAPI",
    "PickListsAPI",
    "PickListOptionsAPI",
    "ProjectsAPI",
    "RelationshipsAPI",
    "TagsAPI",
    "TestCyclesAPI",
    "TestPlansAPI",
    "TestRunsAPI",
    "UsersAPI",
]



"""
Tags API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> tags_api = TagsAPI(client)
    >>> tags = tags_api.get_tags()
"""

import json
import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class TagsAPI:
    client: JamaClient

    resource_path = "tags"

    def __init__(self, client: JamaClient):
        self.client = client

    def get_tags(
        self,
        project_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all tags for the project with the specified id
        Args:
            project: The API ID of the project to fetch tags for.
            allowed_results_per_page: Number of results per page

        Returns: A Json Array that contains all the tag data for the specified
        project.

        """
        req_params = {"project": project_id}
        if params is None:
            params = req_params
        else:
            params.update(req_params)

        return self.client.get_all(
            self.resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )

    def post_tag(
        self,
        name: str,
        project: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Create a new tag in the project with the specified project ID
        Args:
            name: The display name for the tag
            project: The project to create the new tag in

        Returns: the newly created Tag
        """
        body = {"name": name, "project": project}
        headers = {"content-type": "application/json"}
        try:
            response = self.client.post(
                self.resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_tag(
        self,
        tag_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Gets item tag information for a specific item tag id.
        Args:
            item_tag_id: The api id of the item tag to fetch

        Returns: JSON object
        """
        resource_path = f"tags/{tag_id}"
        try:
            response = self.client.get(resource_path, params)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def put_tag(
        self,
        tag_id: int,
        name: str,
        project: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Update an existing tag with the specified tag ID in the project with
        the specified project ID
        Args:
            tag_id: integer API id of the tag
            name: string name of the tag
            project: the project in which to update the tag
        """
        body = {"name": name, "project": project}
        resource_path = "tags/{}".format(tag_id)
        headers = {"content-type": "application/json"}
        try:
            response = self.client.put(
                resource_path,
                params,
                data=json.dumps(body),
                headers=headers,
                **kwargs
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def delete_tag(self, tag_id: int) -> int:
        """
        Deletes a tag with the specified tag ID
        Args:
            tag_id: the api id of a tag

        Returns: The success status code.
        """
        resource_path = f"tags/{tag_id}"
        try:
            response = self.client.delete(resource_path)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return response.status_code

    def get_tag_items(
        self,
        tag_id: int,
        *args,
        params: Optional[dict] = None,
        allowed_results_per_page=DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        **kwargs,
    ):
        """
        Get all items that have the tag with the specified id
        Args:
            allowed_results_per_page: Number of results per page

        Returns: A Json Array containing all items tagged with the specified
        tag id.
        """
        resource_path = f"tags/{tag_id}/items"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page=allowed_results_per_page,
            **kwargs,
        )



"""
Activities API module

Example usage:

    >>> from py_jama_client.client import JamaClient
    >>> client = JamaClient(host=HOST, credentials=(USERNAME, PASSWORD))
    >>> activities_api = ActivitiesAPI(client)
    >>> activities = activities_api.get_activities(project_id=82)
"""

import logging
from typing import Optional

from py_jama_client.client import JamaClient
from py_jama_client.constants import DEFAULT_ALLOWED_RESULTS_PER_PAGE
from py_jama_client.exceptions import APIException, CoreException
from py_jama_client.response import ClientResponse

py_jama_client_logger = logging.getLogger("py_jama_client")


class ActivitiesAPI:
    client: JamaClient

    resource_path = "activities"

    def __init__(self, client: JamaClient) -> None:
        self.client = client

    def get_activities(
        self,
        project_id: int,
        event_type: list[str] = None,
        object_type: list[str] = None,
        item_type: list[int] = None,
        date: list[str] = None,
        delete_events: bool = None,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all activities in the project with the specified ID
        GET /activities/

        Args:
            project_id: Project resource id
            event_type: Event type to filter on. More than one event type can
                be chosen
                Available values : CREATE, BATCH_CREATE, UPDATE, BATCH_UPDATE,
                DELETE, BATCH_DELETE, PUBLIC, BATCH_SUMMARY, COPY, BATCH_COPY,
                MOVE, APPLY, MERGE, CREATE_INSTANCE, EDIT, EDIT_PROJECT,
                REMOVE_INSTANCE, REMOVE
            object_type: Object type to filter on. More than one object type
                can be chosen
                Available values : PROJECT, ITEM, USER, RELATIONSHIP, COMMENT,
                ITEM_TAG, TAG, ITEM_ATTACHMENT, URL, TEST_RESULT, BASELINE,
                CHANGE_REQUEST, REVIEW, REVISION, REVISION_ITEM, TEST_PLAN,
                TEST_CYCLE, TEST_RUN, INTEGRATION, MISCELLANEOUS, CATEGORY,
                CATEGORIZED_ITEM, CATEGORY_PATH
            item_type: ID of item type to filter on. More than one item type
                can be chosen
            date: Filter datetime fields after a single date or within a range
                of values. Provide one or two values in ISO8601 format
                (milliseconds or seconds)
                - "yyyy-MM-dd'T'HH:mm:ss.SSSZ" or "yyyy-MM-dd'T'HH:mm:ssZ"
            delete_events: Get item delete events only
        """
        req_params = {"project": project_id}
        if params is None:
            params = req_params
        else:
            params.update(req_params)

        if event_type is not None:
            params.update({"objectType": object_type})
        if item_type is not None:
            params.update({"objectType": object_type})
        if date is not None:
            params.update({"date": date})
        if delete_events is not None:
            params.update({"delete": delete_events})

        return self.client.get_all(
            self.resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )

    def get_activity(
        self,
        activity_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get the activity with the specified ID
        GET /activities/{activityId}

        Args:
            activity_id: (int) activity resource id
        """
        resource_path = f"{self.resource_path}/{activity_id}"
        try:
            response = self.client.get(resource_path, params, **kwargs)
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_activity_affected_items(
        self,
        activity_id: int,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all items affected by the activity with the specified ID
        GET: /activities/{activityId}/affecteditems

        Args:
            activity_id: (int) activity resource id
        """
        resource_path = f"{self.resource_path}/{activity_id}/affecteditems"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )

    def restore_activity_items(
        self,
        activity_id: int,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Restore item(s) associated with a delete activity.
        POST: /activities/{activityId}/restore

        Args:
            activity_id: (int) activity resource id
        """
        resource_path = (
            f"{self.resource_path}/activities/{activity_id}/restore")
        try:
            response = self.client.post(
                resource_path,
                params,
                **kwargs,
            )
        except CoreException as err:
            py_jama_client_logger.error(err)
            raise APIException(str(err))
        JamaClient.handle_response_status(response)
        return ClientResponse.from_response(response)

    def get_admin_activities(
        self,
        filter_term: str = None,
        project_id: int = None,
        allowed_results_per_page: int = DEFAULT_ALLOWED_RESULTS_PER_PAGE,
        *args,
        params: Optional[dict] = None,
        **kwargs,
    ):
        """
        Get all activities at the admin level
        GET: /activities/adminActivity

        Args:
            filter_term: (str) Filter on the text contents of the activities.
                         - Strings in quotations taken literally.
                         - Multiple values are treated as separate tokens
                                for matching.
            project_id: (int) Filter by Project ID. User must be at least
                        Project Administrator
        """
        if params is None:
            params = {}
        if filter_term is not None:
            params.update({"filterTerm": filter_term})
        if project_id is not None:
            params.update({"projectId": project_id})

        resource_path = f"{self.resource_path}/adminActivity"
        return self.client.get_all(
            resource_path,
            params,
            allowed_results_per_page,
            **kwargs,
        )
